--SkyRide Universal Flight Core
--Made by: Kanderman

M = {}

M.active = false
M.targetAlt = 0
M.thrust = 0
M.yawthr = 0
M.altMode = false
M.targetAltAdj = 0

M.stabs = {}
M.thrusts = {}
M.particleEmitters = {}
M.totalMass = 0

local stabs = {}

local AltAuth = 0

local PitchAuth = 0
local RollAuth = 0
local YawAuth = 0

local lastYaw = 0

local flight_noise

local transform_noise

local ls = false

--PID controllers
local aPID = newPIDParallel(0.5, 0.25, 0.25, -2, 4) --Altitude

local pPID = newPIDParallel(4, 1.25, 2, -5, 5) --Pitch
local rPID = newPIDParallel(4, 0, 1, -5, 5) --Roll

local min = math.min
local max = math.max
local clamp = function(v, mi, ma) return max(min(v, ma), mi)end

local function changeMode()
	if electrics.values.SkyRide_mode == 1 then
		electrics.values.SkyRide_mode = 0
		M.active = false
		controller.mainController.setEngineIgnition(true)
		controller.mainController.setStarter(true)
	elseif electrics.values.SkyRide_mode == 0 and electrics.values.ignition ~= 0 then
		electrics.values.SkyRide_mode = 1
		M.active = true
		controller.mainController.setEngineIgnition(false)
		if M.altMode then
			M.targetAlt = obj:castRayStatic(obj:getFrontPosition(), vec3(0,0,-1), 1000) + 2
			guihooks.message("Terrain follow mode: ACTIVE", 3, "SkyRide")
		else
			M.targetAlt = electrics.values.altitude + 1
			guihooks.message("Terrain follow mode: INACTIVE", 3, "SkyRide")
		end
	else
		guihooks.message("Turn on the ignition to fly", 3, "SkyRide")
		return
	end
end

local function altChange(val)
	M.targetAltAdj = val*5
end


local function hoverMode()
	if M.altMode then
		M.altMode = false
		M.targetAlt = electrics.values.altitude
		guihooks.message("Terrain follow mode: INACTIVE", 3, "SkyRide")
	else
		M.altMode = true
		M.targetAlt = obj:castRayStatic(obj:getFrontPosition(), vec3(0,0,-1), 1000)
		guihooks.message("Terrain follow mode: ACTIVE", 3, "SkyRide")
	end
end

local function update(dt)
	if ls ~= (electrics.values.SkyRide_mode == 1) and v.data.SkyRide_flying_noise then
		if electrics.values.SkyRide_mode == 1 then
			obj:playSFX(flight_noise)
		else
			obj:stopSFX(flight_noise)
		end
		obj:playSFX(transform_noise)
		ls = electrics.values.SkyRide_mode == 1
	end
	if M.active then
		local vPos = obj:getPosition()
		local vVel = obj:getVelocity()
		local vDir = obj:getDirectionVector()
		local vDirR = obj:getDirectionVectorRight()
		local vPosF = obj:getFrontPosition()

		local roll, pitch, yaw = obj:getRollPitchYaw()

		local vVelL = vVel:length()
		local vVelX = vVel:dot(vDir)
		local vVelY = vVel:dot(vDirR)

		local yawVel = (yaw - lastYaw)/dt

		lastYaw = yaw

		local vCog = obj:calcCenterOfGravity(false)

		local vCogL = (vCog - vPos)
		local qx, qy, qz, qw = obj:getRotation()
		vCogL:setRotate(quat(qx, qy, qz, -qw))

		AltAuth = 0

		PitchAuth = 0
		RollAuth = 0
		YawAuth = 0

		for k, stab in pairs(M.stabs) do
			local pos1 = (obj:getNodePositionRelative(stab.id1) - vCogL)
			local pos2 = (obj:getNodePositionRelative(stab.id2) - vCogL)
			local dir = (pos2 - pos1):normalized()
			local mags = pos2:cross(dir)
			local upd = dir:dot(vec3(0,0,-1))

			PitchAuth = PitchAuth + math.abs(mags.x)
			RollAuth = RollAuth + math.abs(mags.y)
			YawAuth = YawAuth + math.abs(mags.z)

			AltAuth = AltAuth + upd

			stab.pos = pos2
			stab.upd = upd
			stab.mags = mags

			stabs[k] = stab
		end

		PitchAuth = PitchAuth/#stabs
		RollAuth = RollAuth/#stabs
		YawAuth = YawAuth/#stabs

		AltAuth = AltAuth/#stabs

		local alt

		if M.altMode then
			alt = obj:castRayStatic(vPosF, vec3(0,0,-1), 5 + M.targetAlt*2)
		else
			alt = vPos.z
		end

		local altAdj = (2 + aPID:get(alt, M.targetAlt, dt)*2)/AltAuth

		local pitchadj = pPID:get(pitch, 0, dt)/PitchAuth
		local rolladj = rPID:get(roll, clamp((vVelY/12)/(1+vVelL/30), -0.6, 0.6), dt)/RollAuth
		local yawadj = (((-electrics.values.steering_input*3) - yawVel)*2)/YawAuth

		for _, stab in pairs(stabs) do
			local force = (altAdj*(AltAuth/stab.upd))*(PitchAuth/math.abs(stab.mags.x))*(RollAuth/math.abs(stab.mags.y))*(YawAuth/math.abs(stab.mags.z))
			force = force + pitchadj/(PitchAuth/stab.mags.x)
			force = force - rolladj/(RollAuth/stab.mags.y)
			force = force - yawadj/(YawAuth/stab.mags.z)

			local clmp = (30/#stabs)/(AltAuth/stab.upd)

			obj:applyForce(stab.id2, stab.id1, clamp(force, -clmp, clmp)*M.totalMass)
		end

		local brake = electrics.values.brake_input

		if vVelX <= 0 then
			brake = -brake
		end

		if electrics.values.gear == -1 then
			M.thrust = (-electrics.values.throttle_input/4 - brake*2)*10
		else
			M.thrust = (electrics.values.throttle_input - brake*2)*10
		end

		local thr = M.thrust/(1+math.abs(vVelX)/80) - (vVelX/35)

		for _, thruster in pairs(M.thrusts) do
			obj:applyForce(thruster.id2, thruster.id1, thr*M.totalMass/#M.thrusts)
		end

		local mul = 1

		if M.altMode and M.targetAlt < 8 then
			if M.targetAlt < 4 then
				mul = 0.25
			else
				mul = 0.5
			end
		end

		M.targetAlt = max(M.targetAlt + M.targetAltAdj*mul*dt, 0)
	end
end

local function updateGFX(dt)
	if electrics.values.SkyRide_mode == 1 then
		for _, jet in pairs(M.particleEmitters) do
			obj:addParticleByNodesRelative(jet.id2, jet.id1, -3, 666, 0.1, 3)
		end
		if v.data.SkyRide_flying_noise then
			obj:setVolume(flight_noise, math.min(electrics.values.airspeed/6, 0.75))
			obj:setPitch(flight_noise, 1 + electrics.values.airspeed/3)
		end
	end
end

M.onPhysicsStep = update
M.updateGFX = updateGFX

local function reset()
	if #M.stabs == 0 then
		M.onPhysicsStep = nop
		M.updateGFX = nop
		return
	end

	if v.data.SkyRide_flying_noise then
		obj:stopSFX(flight_noise)
	end

	electrics.values.SkyRide_mode = 0
	M.active = false

	local totalMass, _, _ = extensions.vehicleEditor_nodes.calculateNodesWeight()
	M.totalMass = totalMass
	
	aPID:reset()
	pPID:reset()
	rPID:reset()

	calcTimer = 0
end

M.onReset = reset

local function init()
	M.stabs = {}

	M.thrusts = {}

	M.particleEmitters = {}
	
	if v.data.SkyRide_thrusters then
		for _, thruster in pairs(v.data.SkyRide_thrusters) do
			if thruster.stab then
				table.insert(M.stabs, thruster)
			else
				table.insert(M.thrusts, thruster)
			end
			if thruster.effects then
				table.insert(M.particleEmitters, thruster)
			end
		end
	end

	if v.data.SkyRide_flying_noise then
		local noisemaker = v.data.SkyRide_flying_noise[0].id
		flight_noise = obj:createSFXSource2(v.data.SkyRide_flying_noise[0].file, "AudioDefaultLoop3D", "flying noise", noisemaker, 0)
		obj:stopSFX(flight_noise)
		transform_noise = obj:createSFXSource2("event:>Vehicle>Hydraulics>Pump_Big", "AudioDefault3D", "Transformation noise", noisemaker, 0)
	end

	enablePhysicsStepHook()

	M.onReset()
end

M.onInit = init

M.changeMode = changeMode
M.hoverMode = hoverMode
M.altChange = altChange

return M